// written by Zheng Gengbin
// program translation for:
// RETURN
// STOP
// IF
// CONTAINS_STMT
// LABEL

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#include <stdio.h>

#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"

// assign.cc
extern void change_oc_expression(SgStatement *s, SgArrayRefExp *array, int counter);

//---------------------------

// STOP
void trans_stop(SgStatement *s)
{
    SgStatement *scope = s->getScopeForDeclare();
    if ( scope->variant() == PROG_HEDR)
    {
        ASSERT(scope->numberOfAttributes(RETURN_LABEL));
	SgLabel *label = (SgLabel *)scope->attributeValue(0,RETURN_LABEL);
	ASSERT(label);
  	SgStatement * gotoStmt = new SgGotoStmt(*label);
        s->insertStmtBefore(*gotoStmt, * s->controlParent());
    	s->deleteStmt();
    }
    else {
        SgSymbol * fname = s->getSymbolInCurrentScope( mixstring, "finalize", NULL);
        SgStatement * afuncall = new SgCallStmt(*fname);
        s->insertStmtBefore(*afuncall, * s->controlParent());
    }
}

//---------------------------

// RETURN
void trans_return(SgStatement *s)
{
	SgStatement *scope = s->getScopeForDeclare();
        ASSERT(scope->numberOfAttributes(RETURN_LABEL));
	SgLabel *label = (SgLabel *)scope->attributeValue(0,RETURN_LABEL);
	ASSERT(label);
  	SgStatement * gotoStmt = new SgGotoStmt(*label);
        s->insertStmtBefore(*gotoStmt, * s->controlParent());
    	s->deleteStmt();
}


// --------------------------
// IF statement

// find if we already have such an expression inserted
int find_same_as_assignref(SgStatement *s , SgExpression *expr)
{
 for (int i=1; i<s->numberOfAttributes(Assign_Ref); i++) {
    	assign_reference * aref = (assign_reference *) (s->attributeValue(i,Assign_Ref));
	if ( (*aref->expression()) == *expr )
		return i;
 }
 return -1;
}

void add_if_reference(SgStatement *s , SgExpression *expr , int &num)
{

  if(expr) {
    switch(expr->variant()) {
    case BOOL_VAL:
    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
      break;
    
    case VAR_REF: 
      break;

    case ARRAY_REF: {
      assign_reference * aref = new assign_reference(right);
      aref->setExpression(expr);
      aref->sameAs = find_same_as_assignref(s,expr);
      s->addAttribute(Assign_Ref,(void *)aref, sizeof(assign_reference));
      break;
    }

    case MINUS_OP:
    case ADD_OP:  
    case MULT_OP:
    case DIV_OP:

    case AND_OP:
    case OR_OP:
    case EQV_OP:
    case NEQV_OP:
      
    case EQ_OP:
    case NOTEQL_OP:
    case LT_OP:
    case LTEQL_OP:
    case GT_OP: 
    case GTEQL_OP: {
      add_if_reference(s, expr->operand(1), num);
      add_if_reference(s, expr->operand(2), num);
      break;
    }

    case NOT_OP: {
      add_if_reference(s, expr->operand(1), num);
      break;
    }

    case FUNC_CALL:
    case INTRIN_FUNC_CALL: {
      SgFunctionRefExp * funExp = (SgFunctionRefExp *) expr;
      for (int i=0; i<funExp->numberOfArgs(); i++) {
	add_if_reference(s, funExp->arg(i), num);
      };
	// collect intrinsic
      collect_intrinsic(s,expr,num);
      num++;
      break;
    }

      // select case
    case RANGE_LIST:
    case RANGE_OP: {
      add_if_reference(s, expr->operand(1), num);
      add_if_reference(s, expr->operand(2), num);
      break;
    }

    default:
      printf("In add_if_reference, unexpected expression on line %d, id: %d, varaint: %d\n",s->lineNumber(), expr->id(),expr->variant());

    };
  };
}


void analyse_if_stmt(SgStatement *s)
{
  int i;
  SgAssignStmt * assign = (SgAssignStmt *) s;

   // counter == 0 means lhs, so skip it
  s->addAttribute(Assign_Ref,NULL, 0);
  int funcnum = 0;
  add_if_reference(assign, assign->expr(0), funcnum);
  
  SgStatement *where = s;
  while (where->childList2(0) && where->childList2(0)->variant() == ELSEIF_NODE)
  { 
	where=where->childList2(0);
  	add_if_reference(assign, where->expr(0), funcnum);
  }

  // with some optimization, detect some none communication cases.
  for (i=1; i<assign->numberOfAttributes(Assign_Ref); i++) 
  {
	assign_reference * aref = (assign_reference *)assign->attributeValue(i, Assign_Ref);
  	SgSymbol * arraysymb  = aref->expression()->symbol();
	if ( expr_no_communication(s, arraysymb) )
      	   aref->setCommunication(NONE);
	else
      	   aref->setCommunication(REMAP);
    }
}

void if_create_section(SgAssignStmt * assign, int counter)
{
  assign_reference * aref = (assign_reference *) ((assign->getAttribute(counter, Assign_Ref))->getAttributeData());
  
  SgArrayRefExp * array = (SgArrayRefExp *) aref->expression();
  SgSymbol * arraysymb  = aref->expression()->symbol();

  {
    SgSymbol * fun = assign->getSymbolInCurrentScope(mixstring, "new_array_section", NULL);
    SgFunctionRefExp * funExp = new SgFunctionRefExp (*fun);

    SgValueExp * zero = new SgValueExp(0);

    funExp->addArg(*zero);

    SgSymbol * dad = assign->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
    SgExpression * dadexp = new SgVarRefExp(dad);

    funExp->addArg(*dadexp);

    SgSymbol * sym=assign->getSymbolInCurrentScope(mixstring, "sdd", NULL, if_level ,counter);
    SgExpression * dadsecExp = new SgVarRefExp (sym);
    
    SgAssignStmt * assignStmt = new SgAssignStmt (*dadsecExp, *funExp);
    
    assign->insertStmtBefore(*assignStmt, * assign->controlParent());
  }

  for (int i=0; i<array->numberOfSubscripts(); i++) {
    if (1) {
      SgSymbol * symb= assign->getSymbolInCurrentScope(mixstring, "set_array_scalar", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);
      {
	SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, "sdd", NULL, if_level ,counter);
	SgExpression * dadsecexp = new SgVarRefExp(symb);

	funcall->addArg(*dadsecexp);
      }
      {
	SgExpression * scalar = ( (SgArrayRefExp *) aref->expression() ) ->subscript(i);
	funcall->addArg(*scalar);
      }
      {
        SgSymbol * dad = assign->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
	SgExpression * dadexp = new SgVarRefExp(dad);
	
	funcall->addArg(*dadexp);
      };
      {
	SgValueExp * dim = new SgValueExp (i+1);
	funcall->addArg(*dim);
      }
      assign->insertStmtBefore(*funcall, *assign->controlParent());

    }
    else {
      printf("It should be a scalar on this dim\n");
    }

  }

  // call delete_array;
  {
    SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, "delete_array", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "sdd", NULL, if_level ,counter);
    SgExpression * baseexpr=new SgExpression(VAR_REF, NULL, NULL, base);
    funcall->addArg(*baseexpr);
    
    assign->insertStmtAfter(*funcall, *assign->controlParent());
  }
}

void if_create_new(SgAssignStmt * assign, int counter)
{
//  assign_reference * leftref = (assign_reference *) ((assign->getAttribute(0, Assign_Ref))->getAttributeData());
//  SgSymbol * leftsymb = leftref->expression()->symbol();

  assign_reference * aref = (assign_reference *) ((assign->getAttribute(counter, Assign_Ref))->getAttributeData());
  
  SgArrayRefExp * array = (SgArrayRefExp *) aref->expression();
  SgSymbol * arraysymb  = aref->expression()->symbol();

  int storage = getArrayStorage(arraysymb);

  {
/*
    SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, "new_array_copy", NULL);
    SgFunctionRefExp * funcall = new SgFunctionRefExp (*symb);

    SgSymbol * stack = assign->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysymb->type()), NULL); 
    SgExpression * stackExp = new SgExpression(VAR_REF, NULL, NULL, stack);
    funcall->addArg(*stackExp);    
						       
    SgSymbol * secDad = assign->getSymbolInCurrentScope(mixstring, "sdd", NULL,if_level ,counter);
    SgExpression * secDadExp=new SgExpression(VAR_REF, NULL, NULL, secDad);
    funcall->addArg(*secDadExp);

    SgSymbol * dadBase = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, if_level ,counter);
    SgExpression * dadBaseExp=new SgExpression(VAR_REF, NULL, NULL, dadBase);

    SgAssignStmt * assignStmt = new SgAssignStmt(*dadBaseExp, *funcall);
  
    assign->insertStmtBefore(*assignStmt, *assign->controlParent());
*/
	// by zheng.gb pcrc_new_array_copy ==> pcrc_new_array_data  1997/9/15
   
    SgSymbol * dadBase = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, if_level,counter);

    SgStatement *assignStmt = make_new_array_data(assign,dadBase,arraysymb->type(),
                                                 storage, 0, NULL);
    assign->insertStmtBefore(*assignStmt, *assign->controlParent());
  }

  // allocate memory;
  SgSymbol * mHandle;
  {
    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tbs", NULL,if_level , counter);
    SgSymbol * dadbase = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, if_level ,counter);

#ifdef MEM_ALLOC
    	mHandle = assign->getSymbolInCurrentScope(mixstring, "mh", NULL, if_level, counter);
	SgStatement *alloc=make_stack_alloc(assign,arraysymb->type(),base,dadbase,mHandle);
#else
	SgStatement *alloc=make_alloc(assign,arraysymb->type(),base,dadbase);
#endif

    assign->insertStmtBefore(*alloc, *assign->controlParent());
  }

  // reset base;
  {
    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL,if_level , counter);
    SgSymbol * basesym = assign->getSymbolInCurrentScope(mixstring, "tbs", NULL,if_level ,counter);

    SgStatement *funcall=make_reset_array_base(assign,base,arraysymb->type(),basesym,mHandle);

    assign->insertStmtBefore(*funcall, *assign->controlParent());
  }

  // call delete_array;
  {
    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL,if_level , counter);
    SgStatement *funcall = make_delete_array(assign,base);
    
    assign->insertStmtAfter(*funcall, *assign->controlParent());
  }

  // call real free;
  {
#ifdef MEM_ALLOC
    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "mh", NULL, if_level,counter);
#else
    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tbs", NULL, if_level ,counter);
#endif

    SgStatement * funcall = make_free(assign,arraysymb->type(),base);
    
    assign->insertStmtAfter(*funcall, *assign->controlParent());
  }

}

void if_add_remap(SgAssignStmt * assign, int counter)
{

  SgSymbol *dstsymb=assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, if_level,counter);
  SgSymbol *srcsymb=assign->getSymbolInCurrentScope(mixstring, "sdd", NULL ,if_level, counter);
  SgStatement *funcall=make_remap(assign,dstsymb,srcsymb);

  assign->insertStmtBefore(*funcall, *assign->controlParent());
}

void if_change_assign_reference(SgAssignStmt * assign, int counter )
{
  assign_reference * aref = (assign_reference *) assign->attributeValue(counter, Assign_Ref);

  int num = aref->sameAs;
  if ( num == -1 ) num = counter;

  SgArrayRefExp * array = (SgArrayRefExp *) aref->expression();
  SgSymbol * arraysymb  = aref->expression()->symbol();

#ifdef MEM_ALLOC
  SgSymbol * stack = assign->getSymbolInCurrentScope(stackNameOfType(arraysymb->type()),"mb", NULL); 
#else
  SgSymbol * stack = assign->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysymb->type()), NULL); 
#endif
  SgArrayRefExp * stackExp = new SgArrayRefExp(* stack);
  
  SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tbs", NULL, if_level,num);
  SgExpression * baseexp=new SgVarRefExp(base);
  stackExp->addSubscript(*baseexp);

  array->replace(*stackExp);
}

// counter is the sequnce number , num is the sequce num of expression to replace
void trans_if_reference(SgStatement *s , int counter )
{
  assign_reference * aref = (assign_reference *)s->attributeValue(counter, Assign_Ref);

  // use the way we deal with the assign statement !
  switch (aref->getCommunication())
  {
  case REMAP: {
	if ( aref->sameAs == -1 )
	{
      	  if_create_section((SgAssignStmt *)s, counter);
      	  if_create_new((SgAssignStmt *)s, counter);
      	  if_add_remap((SgAssignStmt *)s, counter);
	}
      	if_change_assign_reference((SgAssignStmt *)s, counter);
    };
    break;
  case NONE: {
  	SgArrayRefExp * array = isSgArrayRefExp(aref->expression());
	ASSERT(array);
	analyse_array_ref_in_exp(s,array, s);
    }
    break;
  default:
    ASSERT(0);
  }
}

void trans_if_stmt(SgStatement *s)
{

  int numOfAttr = s->numberOfAttributes(Assign_Ref);

    // start from 1, because 0 always means left hand side
  for (int i=1; i< numOfAttr; i++) {
      trans_if_reference(s, i );
  }


    // transforam intrinsic functions
  analyse_intrinsic_list(s,s);
}


//----------------------------------

// SELECT CASE
void analyse_select_stmt(SgStatement *s)
{
  int i;
  SgAssignStmt * assign = (SgAssignStmt *) s;

  int funcnum = 0;
  add_if_reference(assign, assign->expr(0), funcnum);
  
/*
 case exprlist must be constant
  int num = s->numberOfChildrenList1();
  for (i=0; i< num; i++) 
	if (s->childList1(i)->variant() == CASE_NODE)
  	   add_if_reference(assign, s->childList1(i)->expr(0), funcnum);
*/

  // set the communication type as REMAP now , without further optimization
  for (i=0; i<assign->numberOfAttributes(Assign_Ref); i++) 
  {
	assign_reference * aref = (assign_reference *) (assign->attributeValue(i, Assign_Ref));
      	aref->setCommunication(REMAP);
    }
}

//-------------------------------------

// CONTAINS_STMT
// only called if no F90 compiler
void extract_contains(SgStatement *scope)
{
    int in_contain = FALSE;
    int org_numberOfChildren=scope->numberOfChildrenList1();
    SgStatement * * children = new SgStatement * [org_numberOfChildren];
    int i; 

printf("Warnign: extract subroutines from contains block cannot always correct.\n");

    for (i=0; i<org_numberOfChildren; i++)
	children[i]=scope->childList1(i);

    for (i=0; i<org_numberOfChildren ; i++)
    {
	SgStatement *stmt = children[i];
	if ( stmt->variant() == CONTAINS_STMT ) { 
	    in_contain=TRUE; 
            stmt->deleteStmt();
	    continue; 
	}
 	if (in_contain == FALSE) continue;
        if (stmt->variant() == PROC_HEDR || stmt->variant() == FUNC_HEDR)
	{
	    SgStatement *newHeader = stmt->copyPtr();
	    scope->insertStmtAfter( *newHeader, *scope->controlParent());
            stmt->deleteStmt();
if (!silentFlag)
printf("extract subroutines in contains successfully. \n");
	}
    }
    delete [] children;
}

//----------------------------------

// keep LABEL
void process_label(SgStatement *s)
{
    // leave this cases unchanged
  switch (s->variant()) {
  case FORMAT_STAT:
  case CONTINUE_NODE:
  case ENDDO_STMT:
	return;
  }

  if (s->hasLabel()) {
    SgLabel * label = s->label();

    SgStatement * continueStmt = new SgStatement(CONTINUE_NODE);

    continueStmt->setLabel(*label);

    s->insertStmtBefore(*continueStmt, *s->controlParent());

    s->setLabel(NULL);
  }
}



